﻿using SharpMap.Data.Providers;
using SharpMap.Geometries;
using SharpMap.Layers;
using System;
using System.Collections.Generic;
using System.Data.Spatial;
using System.Drawing;
using System.Globalization;
using System.Linq;

namespace DevartSharpMapDemo {

  public class OracleToSharpMapLoader {

    private SharpMap.Map map;

    public OracleToSharpMapLoader() {

      this.map = new SharpMap.Map(new Size(1500, 750));
      this.map.BackColor = Color.White;
    }

    #region France map

    public Image LoadFranceMap(out List<string> cityNames) {

      // France regions
      VectorLayer mapLayer = new VectorLayer("France regions");
      using (var ctx = new Model()) {
        var regions = ctx.Regions
          .Select(row => row.Geometry)
          .AsEnumerable()
          .Select(a => (Geometry)a.ProviderValue)
          .ToList();
        mapLayer.DataSource = new GeometryProvider(regions);
      }
      mapLayer.Style.Fill = new SolidBrush(Color.FromArgb(90, 200, 255));
      mapLayer.Style.EnableOutline = true;
      mapLayer.Style.Outline = Pens.White;
      map.Layers.Add(mapLayer);

      // Cities
      VectorLayer cityLayer = new VectorLayer("Cities");
      using (var ctx = new Model()) {
        var cities = ctx.Cities
          .OrderBy(city => city.Name)
          .ToList();
        cityNames = new List<string>(cities.Count);

        LabelTable table = new LabelTable();
        foreach (var city in cities) {
          table.AddRow(city.Geometry, city.Name);
          cityNames.Add(city.Name);          
        }
        cityLayer.DataSource = new GeometryFeatureProvider(table);
      }
      cityLayer.Style.PointColor = new SolidBrush(Color.MediumOrchid);
      cityLayer.Style.PointSize = 8;
      map.Layers.Add(cityLayer);

      // City labels
      LabelLayer labelLayer = new LabelLayer("City Names");
      labelLayer.DataSource = cityLayer.DataSource;
      labelLayer.LabelColumn = LabelTable.LabelColumnName;
      labelLayer.Style.CollisionDetection = true;
      labelLayer.Style.CollisionBuffer = new SizeF(10, 10);
      labelLayer.Style.MaxVisible = 200;
      labelLayer.MultipartGeometryBehaviour = LabelLayer.MultipartGeometryBehaviourEnum.Largest;
      labelLayer.Style.Font = new Font(FontFamily.GenericSansSerif, 12);
      labelLayer.Style.VerticalAlignment = SharpMap.Styles.LabelStyle.VerticalAlignmentEnum.Bottom;
      map.Layers.Add(labelLayer);

      // Refresh
      return GetMapImage();
    }

    #endregion

    #region Find distance

    public Image FindDistance(string firstCityName, string secondCityName) {

      ClearQueryLayers();

      using (var ctx = new Model()) {

        List<DistanceResult> results;
        if (secondCityName != "all cities") {
          var query = from city in ctx.Cities
                      where city.Name == firstCityName
                      let secondCity = ctx.Cities
                        .Where(secondCity => secondCity.Name == secondCityName)
                        .Select(secondCity => secondCity.Geometry)
                        .FirstOrDefault()
                      select new DistanceResult()
                      {
                        FirstCity = city.Geometry,
                        SecondCity = secondCity,
                        Distance = city.Geometry.Distance(secondCity)
                      };
          results = query.ToList();
        }
        else {
          var query = from city in ctx.Cities
                      join secondCity in ctx.Cities on city.Name equals firstCityName
                      select new DistanceResult()
                      {
                        FirstCity = city.Geometry,
                        SecondCity = secondCity.Geometry,
                        Distance = city.Geometry.Distance(secondCity.Geometry)
                      };
          results = query.ToList();
        }

        LabelTable table = new LabelTable();
        foreach (var item in results) {
          Geometry line = new LineString(new SharpMap.Geometries.Point[] {
                (SharpMap.Geometries.Point)item.FirstCity.ProviderValue,
                (SharpMap.Geometries.Point)item.SecondCity.ProviderValue
              });
          string label = ((int)Math.Round(item.Distance.GetValueOrDefault())).ToString(CultureInfo.InvariantCulture) + " km";
          table.AddRow(line, label);
        }
        VectorLayer queryDataLayer = new VectorLayer(QueryDataLayerName);
        queryDataLayer.DataSource = new GeometryFeatureProvider(table);
        queryDataLayer.Style.EnableOutline = true;
        queryDataLayer.Style.Outline = Pens.LightPink;
        map.Layers.Add(queryDataLayer);

        LabelLayer labelLayer = new LabelLayer(queryDataLayer.LayerName + " : Distance labels");
        labelLayer.DataSource = queryDataLayer.DataSource;
        labelLayer.LabelColumn = LabelTable.LabelColumnName;
        labelLayer.Style.CollisionDetection = true;
        labelLayer.Style.CollisionBuffer = new SizeF(10, 10);
        labelLayer.Style.MaxVisible = 200;
        labelLayer.Style.Font = new Font(FontFamily.GenericSansSerif, 14);
        labelLayer.Style.ForeColor = Color.Red;
        labelLayer.Style.VerticalAlignment = SharpMap.Styles.LabelStyle.VerticalAlignmentEnum.Top;
        map.Layers.Add(labelLayer);
      }

      return GetMapImage();
    }

    private class DistanceResult {

      public DbGeography FirstCity { get; set; }
      public DbGeography SecondCity { get; set; }
      public double? Distance { get; set; }
    }

    #endregion

    #region Greatests areas

    public Image FindGreatestRegions(int count, bool useDifferentColors, bool showLabels) {

      ClearQueryLayers();

      if (!showLabels)
        ShowAreasWithoutCaptions(count, useDifferentColors);
      else
        ShowAreasWithCaptions(count, useDifferentColors);

      return GetMapImage();
    }

    private void ShowAreasWithoutCaptions(int count, bool useDifferentColors) {

      using (var ctx = new Model()) {        

        var query = ctx.Regions
          .OrderByDescending(region => region.Geometry.Area)
          .Take(count)
          .Select(region => region.Geometry);
        var results = query.ToList();

        if (!useDifferentColors) {
          VectorLayer queryDataLayer = new VectorLayer(QueryDataLayerName);
          queryDataLayer.DataSource = new GeometryProvider(results.Select(a => (Geometry)a.ProviderValue).ToArray());
          AddGreatestRegionQueryLayer(queryDataLayer, new SolidBrush(Color.FromArgb(0, 127, 255)));
        }
        else {
          int rDelta = 255 / results.Count;
          int gDelta = 127 / results.Count;
          for (int i = 0; i < results.Count; i++) {
            VectorLayer queryDataLayer = new VectorLayer(QueryDataLayerName + " " + i.ToString());
            queryDataLayer.DataSource = new GeometryProvider(new Geometry[] { (Geometry)results[i].ProviderValue });
            AddGreatestRegionQueryLayer(queryDataLayer, new SolidBrush(Color.FromArgb(i * rDelta, 127 + i * gDelta, 255)));
          }
        }
      }
    }

    private void ShowAreasWithCaptions(int count, bool useDifferentColors) {

      using (var ctx = new Model()) {

        var query = ctx.Regions
          .Select(region => new
          {
            Geometry = region.Geometry,
            Area = region.Geometry.Area
          })
          .OrderByDescending(row => row.Area)
          .Take(count);
          
        var results = query.ToList();

        if (!useDifferentColors) {
          VectorLayer queryDataLayer = new VectorLayer(QueryDataLayerName);

          LabelTable table = new LabelTable();
          foreach (var region in results)
            table.AddRow(region.Geometry, GetSquareKmLabel(region.Area));
          queryDataLayer.DataSource = new GeometryFeatureProvider(table);

          AddGreatestRegionQueryLayer(queryDataLayer, new SolidBrush(Color.FromArgb(0, 127, 255)));
          AddLabelLayer(queryDataLayer);
        }
        else {
          int rDelta = 255 / results.Count;
          int gDelta = 127 / results.Count;
          for (int i = 0; i < results.Count; i++) {
            VectorLayer queryDataLayer = new VectorLayer(QueryDataLayerName + " " + i.ToString());

            LabelTable table = new LabelTable();
            table.AddRow(results[i].Geometry, GetSquareKmLabel(results[i].Area));
            queryDataLayer.DataSource = new GeometryFeatureProvider(table);

            AddGreatestRegionQueryLayer(queryDataLayer, new SolidBrush(Color.FromArgb(i * rDelta, 127 + i * gDelta, 255)));
            AddLabelLayer(queryDataLayer);
          }
        }
      }
    }

    private string GetSquareKmLabel(double? km) {

      return ((int)Math.Round(km.GetValueOrDefault())).ToString(CultureInfo.InvariantCulture) + " km²";
    }

    private void AddGreatestRegionQueryLayer(VectorLayer layer, Brush brush) {

      map.Layers.Insert(1, layer);
      layer.Style.Fill = brush;
      layer.Style.EnableOutline = true;
      layer.Style.Outline = Pens.Blue;
    }

    private void AddLabelLayer(VectorLayer layer) {

      LabelLayer labelLayer = new LabelLayer(layer.LayerName + " : Labels");
      labelLayer.DataSource = layer.DataSource;
      labelLayer.LabelColumn = LabelTable.LabelColumnName;
      labelLayer.Style.CollisionDetection = true;
      labelLayer.Style.CollisionBuffer = new SizeF(10, 10);
      labelLayer.Style.MaxVisible = 200;
      labelLayer.MultipartGeometryBehaviour = LabelLayer.MultipartGeometryBehaviourEnum.Largest;
      labelLayer.Style.Font = new Font(FontFamily.GenericSansSerif, 12);
      labelLayer.Style.ForeColor = Color.Red;
      map.Layers.Add(labelLayer);
    }

    #endregion

    #region Helper methods

    private Image GetMapImage() {

      map.ZoomToExtents();
      return map.GetMap();
    }

    private void ClearQueryLayers() {

      for (int i = map.Layers.Count - 1; i >= 0; i--) {
        var layer = map.Layers[i];
        if (layer.LayerName.StartsWith(QueryDataLayerName))
          map.Layers.Remove(layer);
      }
    }

    #endregion

    private const string QueryDataLayerName = "Query data";

  }

}
